/*
 * Copyright (C) 2004 Jesus Gimenez, Lluis Marquez and Senen Moya
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "list.h"
#include "hash.h"
#include "common.h"
#include "dict.h"

struct stat_t{
	char key[100];
	int hits;
	int trials;
	int mft;
};

hash_t stat_Amb_Level,stat_Class_Amb,stat_POS;
int report_type = 1;
int numAmbLevel=0,numAmbClass=0;

/******************************************************************/

void printHelp()
{
fprintf(stderr,"Usage : SVMTeval.perl [mode] <model> <gold> <pred>\n\n");
fprintf(stderr,"\t- mode:\t0 - complete report (everything)\n");
fprintf(stderr,"\t\t1 - overall accuracy only [default]\n");
fprintf(stderr,"\t\t2 - accuracy of known vs unknown words\n");
fprintf(stderr,"\t\t3 - accuracy per level of ambiguity\n");
fprintf(stderr,"\t\t4 - accuracy per kind of ambiguity\n");
fprintf(stderr,"\t\t5 - accuracy per part-of-speech\n");
fprintf(stderr,"\t- model: model location (path + name)\n");
fprintf(stderr,"\t- gold:  correct tagging file\n");
fprintf(stderr,"\t- pred:  predicted tagging file\n\n");
fprintf(stderr,"Example : SVMTeval.perl WSJTP WSJTP.IN WSJTP.OUT\n");
}

/******************************************************************/

void printHashStats(hash_t *tptr, int put_eol, char *column_name)
{
	fprintf(stderr,"%s\tHITS\t\tTRIALS\t\tACCURACY\t\tMFT-ACCURACY\n",column_name);
	fprintf(stderr,"* ------------------------------------------------------------------------- \n");
	hash_node_t *node, *last;
	int i;

	char c='\0';
	if (put_eol==TRUE) c='\n';

	for (i=0; i<tptr->size; i++)
	{
		node = tptr->bucket[i];
		while (node != NULL)
		{
			last = node;
			node = node->next;
			stat_t *s = (stat_t *)last->data;
			fprintf(stderr,"%s%c\t%d\t/\t%d\t=\t%.4f %\t\t%.4f %\n",s->key,c,s->hits,s->trials, 100*((float)s->hits/s->trials), 100*((float)s->mft/s->trials) );
		}
	}

}

void printKnownVsUnknown(int knownAmb, int knownUnamb, int unknown,int unkHits, int knownHitsAmb,int knownHitsUnamb)
{
int knownHits = knownHitsAmb + knownHitsUnamb;
int known = knownAmb + knownUnamb;

fprintf(stderr,"* ================= KNOWN vs UNKNOWN WORDS ================================\n");
fprintf(stderr,"\tHITS\t\tTRIALS\t\tACCURACY\n");
fprintf(stderr,"* -------------------------------------------------------------------------\n");
fprintf(stderr,"* ======= known ===========================================================\n");
fprintf(stderr,"\t%d\t/\t%d\t=\t%.4f %\n",knownHits,known,100*((float)knownHits/known));
fprintf(stderr,"-------- known unambiguous words ------------------------------------------\n");
fprintf(stderr,"\t%d\t/\t%d\t=\t%.4f %\n",knownHitsUnamb,knownUnamb,100*((float)knownHitsUnamb/knownUnamb));
fprintf(stderr,"-------- known ambiguous words --------------------------------------------\n");
fprintf(stderr,"\t%d\t/\t%d\t=\t%.4f %\n",knownHitsAmb,knownAmb,100*((float)knownHitsAmb/knownAmb));
fprintf(stderr,"* ======= unknown =========================================================\n");
fprintf(stderr,"\t%d\t/\t%d\t=\t%.4f %\n",unkHits,unknown,100*((float)unkHits/unknown));
fprintf(stderr,"* =========================================================================\n");
}

void printTaggingSumary(int known,int unknown,int ambiguous,int well,int wellMFT)
{
fprintf(stderr,"* ================= TAGGING SUMMARY =======================================\n");
fprintf(stderr,"#WORDS\t\t = %d\n",known+unknown);
fprintf(stderr,"#KNOWN\t\t = %d\t/\t%d\t--> (%.4f %)\n",known,known+unknown,100*((float)known/(known+unknown)));
fprintf(stderr,"#UNKNOWN\t = %d\t/\t%d\t--> (%.4f %)\n",unknown,known+unknown,100*((float)unknown/(known+unknown)));
fprintf(stderr,"#AMBIGUOUS\t = %d\t/\t%d\t--> (%.4f %)\n",ambiguous,known+unknown,100*((float)ambiguous/(known+unknown)));
fprintf(stderr,"#MFT baseline\t = %d\t/\t%d\t--> (%.4f %)\n",wellMFT,known+unknown,100*((float)wellMFT/(known+unknown)));
}

void printOverallAccuracy(int total, int well, int wellMFT, float pAmb)
{
fprintf(stderr,"* ================= OVERALL ACCURACY ======================================\n");
fprintf(stderr,"\tHITS\t\tTRIALS\t\tACCURACY\tMFT-baseline\n");
fprintf(stderr,"* -------------------------------------------------------------------------\n");
fprintf(stderr,"\t%d\t/\t%d\t=\t%.4f\t\t%.4f%\n",well,total,100*((float)well/total),100*((float)wellMFT/total));
fprintf(stderr,"* =========================================================================\n");
fprintf(stderr,"\tAmbiguity Average for Ambiguous Known words = %.5f POS/word\n",pAmb);
fprintf(stderr,"* =========================================================================\n");
}


void printStatsByLevel(hash_t *h)
{
fprintf(stderr,"* ================= ACCURACY PER LEVEL OF AMBIGUITY =======================\n");
fprintf(stderr,"#CLASSES = %d\n",numAmbLevel);
fprintf(stderr,"* =========================================================================\n");
printHashStats(h,FALSE,"LEVEL");
}


void printStatsByAmbiguityClass(hash_t *h)
{
fprintf(stderr,"* ================= ACCURACY PER CLASS OF AMBIGUITY =======================\n");
fprintf(stderr,"#CLASSES = %d\n",numAmbClass);
fprintf(stderr,"* =========================================================================\n");
printHashStats(h,TRUE,"CLASS");
}

void printStatsByPOS(hash_t *h)
{
fprintf(stderr,"* =================== ACCURACY PER PART-OF_SPEECH =========================\n");
printHashStats(h,FALSE,"POS");
}

/******************************************************************/

/******************************************************************/

void addStatsToHash(hash_t *h,char *key,int is_hit, int is_mft)
{
	int p = hash_lookup(h,key);
	if (p!=HASH_FAIL)
	{
		stat_t *s = (stat_t *)p;
		strcpy(s->key,key);
		if (is_hit==TRUE) s->hits++;
		if (is_mft==TRUE) s->mft++;
		s->trials++;
	}
	else
	{
		if (report_type==3 || report_type==0) numAmbLevel++;
		if (report_type==4 || report_type==0) numAmbClass++;
		stat_t *s = new stat_t;
		strcpy(s->key,key);
		s->trials=1;
		s->hits=0;
		s->mft=0;
		if (is_hit==TRUE) s->hits++;
		if (is_mft==TRUE) s->mft++;
		hash_insert(h,s->key,(int) s);
	}
}

/******************************************************************/

void simpleReport(dictionary *d,FILE *gold, FILE *pred)
{
	char *mft,wrd1[150],wrd2[150],pos1[5],pos2[5];
	int totalWords=0, well = 0,known=0,unknown=0,wellMFT=0;
	int ambiguous=0,unambiguous=0;
	int unkHits=0,knownHitsAmb=0,knownHitsUnamb=0;
	int ret1=0,ret2=0,ret3=0,ret4=0;
	int is_mft=FALSE,is_hit=FALSE;
	int contAmbiguities=0;

	while (!feof(gold) && !feof(pred))
	{
		is_mft=FALSE;
		is_hit=FALSE;
		ret1 = readTo(gold,' ','\n',wrd1);
		if (wrd1[0]=='#' && wrd1[1]=='#')
		  {
		    char garbage[512];
		    fgets(garbage,512,gold);
		    ret1 = readTo(gold,' ','\n',wrd1);
		  }
		ret2 = readTo(gold,' ','\n',pos1);
	
		ret3 = readTo(pred,' ','\n',wrd2);
		if (wrd2[0]=='#' && wrd2[1]=='#')
		  {
		    char garbage[512];
		    fgets(garbage,512,pred);
		    ret3 = readTo(pred,' ','\n',wrd2);
		  }
		ret4 = readTo(pred,' ','\n',pos2);

		if (ret1>=0 && ret2 >=0 && ret3>=0 && ret4>=0)
		{
			int w = d->getElement(wrd1);
			int numMaybe;
			if (w!=HASH_FAIL)
			{
				known++;
				numMaybe = d->getElementNumMaybe(w);
				
				if (numMaybe>1) 
				    ambiguous++;
				else unambiguous++;

				contAmbiguities += numMaybe;

				mft = d->getMFT(w);
				if (strcmp(mft,pos1)==0)
				{  is_mft = TRUE;
				   wellMFT++;
				}
				delete mft;
			}
			else unknown++;

			if (strcmp(wrd1,wrd2)==0 && strcmp(pos1,pos2)==0)
			{
			 well++;
			 is_hit=TRUE;
			 if (w==HASH_FAIL) unkHits++;
			 else if (numMaybe>1) knownHitsAmb++;
			 else knownHitsUnamb++;
			}

			if (report_type==3 || report_type==0)
			{
			  char level[4];
			  if (w!=HASH_FAIL) sprintf(level,"%d",numMaybe);
			  else sprintf(level,"UNKOWN");
			  addStatsToHash(&stat_Amb_Level,level,is_hit,is_mft);
			}
			if (report_type==4 || report_type==0)
			{
			  char *ambClass = d->getAmbiguityClass(w);
			  addStatsToHash(&stat_Class_Amb,ambClass,is_hit,is_mft);
			  delete ambClass;
			}
			if (report_type==5 || report_type==0)
			{
			  addStatsToHash(&stat_POS,pos2,is_hit,is_mft);
			}

			showProcessDone(totalWords,2000,FALSE,"words");

			if (ret2>0) readTo(gold,'\n', '\n',wrd1);
			if (ret4>0) readTo(pred,'\n', '\n',wrd2);
			totalWords++;
		}

	}

  showProcessDone(totalWords,2000,TRUE,"words");
  printTaggingSumary(known,unknown,ambiguous,well,wellMFT);
  if (report_type==2 || report_type==0)	printKnownVsUnknown(ambiguous, unambiguous, unknown,unkHits,knownHitsAmb,knownHitsUnamb);
  if (report_type==3 || report_type==0) printStatsByLevel(&stat_Amb_Level);
  if (report_type==4 || report_type==0) printStatsByAmbiguityClass(&stat_Class_Amb);
  if (report_type==5 || report_type==0) printStatsByPOS(&stat_POS);

  float porcentageAmbiguedad = (float)contAmbiguities/ (float) known;
  printOverallAccuracy(unknown+known,well,wellMFT,porcentageAmbiguedad);
}


/************************************************************************/

//opciones
//nombre del modelo
//nombre del fichero correcto
//nombre del fichero de predicciones
int main(int argc, char *argv[])
{
	int i = 0;
	if (argc<4)
	{
		fprintf(stderr,"Waiting 3 or more parameters\n");
		printHelp();
		exit(0);
	}
	if (argc>4)
	{
		report_type = atoi(argv[1]);
		i=1;
		if (report_type>5 || report_type<0)
		{
			printHelp();
			exit(0);
		}
	}

	fprintf(stderr,"* ========================= SVMTeval report ==============================\n");
	fprintf(stderr,"* model     = [%s]\n* testset   = [%s]\n* predicted = [%s]\n",argv[1+i],argv[2+i],argv[3+i]);
	fprintf(stderr,"* ========================================================================\n");

	hash_init(&stat_Amb_Level,10);
	hash_init(&stat_Class_Amb,100);
	hash_init(&stat_POS,100);

	char name[200];
	sprintf(name,"%s.DICT",argv[1+i]);
	FILE *gold = openFile(argv[2+i],"rt");
	FILE *pred = openFile(argv[3+i],"rt");
	dictionary *d  = new dictionary(name);
	simpleReport(d,gold,pred);
}
